import os
import yaml
import numpy as np
import pandas as pd
from tqdm import tqdm
from datetime import datetime
from typing import Optional
from sqlalchemy import create_engine, text
from sqlalchemy.engine import URL
from sqlalchemy.exc import ResourceClosedError

def load_db_config(yaml_file_name: str, database: str, custom_path: Optional[str] = None) -> dict:
    """加载数据库配置"""
    file_path = os.path.join(custom_path, yaml_file_name) if custom_path else yaml_file_name
    with open(file_path, 'r') as stream:
        config = yaml.safe_load(stream)
    return config.get(database, {})

def connect_to_db(cfg: dict):
    """创建数据库连接"""
    try:
        url = URL.create(
            drivername='mysql+mysqldb',
            username=cfg['user'],
            password=cfg['password'],
            host=cfg['host'],
            port=cfg['port'],
            database=cfg['database']
        )
        engine = create_engine(url)
        return engine
    except KeyError as e:
        raise ValueError(f"配置中缺少必要的键：{e}")

def clean_dataframe(df: pd.DataFrame):
    """清理DataFrame中的无穷值"""
    numeric_cols = df.select_dtypes(include=[np.number]).columns
    if len(numeric_cols) > 0:
        inf_values = df[numeric_cols].isin([np.inf, -np.inf]).any()
        cols_with_inf = inf_values[inf_values].index.tolist()
        if cols_with_inf:
            df.replace([np.inf, -np.inf], np.nan, inplace=True)
            print(f"警告：以下列包含无穷值，已替换为NaN：{cols_with_inf}")

def execute_sql(engine, sql_statement: str):
    """执行SQL语句"""
    try:
        with engine.begin() as conn:
            conn.execute(text(sql_statement))
    except Exception as e:
        raise ValueError(f"执行SQL时发生错误：{e}")

def update(
        raw_df: pd.DataFrame,
        database: str,
        table: str,
        yaml_file_name: str = 'cfg.yaml',
        clause: Optional[str] = None,
        date_col: Optional[str] = None,
        custom_path: Optional[str] = None
    ):
    """更新数据库表中的数据"""
    try:
        df = raw_df.copy()
        if df.empty:
            raise ValueError("导入的数据集为空。")

        cfg = load_db_config(yaml_file_name, database, custom_path)
        engine = connect_to_db(cfg)
        clean_dataframe(df)

        if clause or date_col:
            # 删除满足条件的数据
            conditions = []
            if clause:
                conditions.append(clause)
            if date_col:
                start_date, end_date = df[date_col].min(), df[date_col].max()
                conditions.append(f"{date_col} BETWEEN '{start_date}' AND '{end_date}'")
            delete_sql = f"DELETE FROM {table} WHERE {' AND '.join(conditions)}"
            execute_sql(engine, delete_sql)

        # 使用chunksize参数上传数据
        df.to_sql(table, engine, if_exists='append', index=False, chunksize=1000, method='multi')
        print_action_result(table, "更新", df, date_col)

    except Exception as e:
        raise ValueError(f"更新过程发生错误：{e}")

def print_action_result(table: str, action: str, df: pd.DataFrame, df_date_col: Optional[str] = None):
    """打印操作结果"""
    current_time = datetime.now().strftime('%Y-%m-%d %H:%M')
    if df_date_col:
        date_range = f"{df[df_date_col].min()} 至 {df[df_date_col].max()}"
        print(f"{table} 数据已{action}：{current_time}\n数据日期范围：{date_range}")
    else:
        print(f"{table} 数据已{action}：{current_time}")

def sql_query(
        database: str,
        sql: str,
        yaml_file_name: str='cfg.yaml',
        custom_path: Optional[str] = None
    ) -> Optional[pd.DataFrame]:
    """执行SQL查询"""
    cfg = load_db_config(yaml_file_name, database, custom_path)
    engine = connect_to_db(cfg)
    try:
        if sql.strip().upper().startswith("SELECT"):
            df = pd.read_sql(sql, engine)
            return df
        else:
            execute_sql(engine, sql)
            print('操作完成。')
    except ResourceClosedError:
        print('查询完成，但没有返回任何数据。')
    except Exception as e:
        raise ValueError(f"执行SQL查询时发生错误：{e}")



import smtplib
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.application import MIMEApplication
import pandas as pd
import glob
import os
from tqdm import tqdm
import chardet

def send_email(subject:str, body:str, send_email_address:str, send_email_password:str, receive_email_address:str, attachment_path=None, attachment_list=None, smtp_address='smtp.feishu.cn', smtp_port=465):
    # Create a multipart message
    msg = MIMEMultipart()
    msg['From'] = send_email_address
    msg['To'] = receive_email_address
    msg['Subject'] = subject

    # Add body text
    msg.attach(MIMEText(body, 'plain'))
    if attachment_path is not None and attachment_list is not None:
        # Add multiple attachments
        attachment_paths = [os.path.join(attachment_path, attachment) for attachment in attachment_list]
        for each_path in attachment_paths:
            attachment_filename = each_path.split("/")[-1]
            with open(each_path, "rb") as f:
                attach = MIMEApplication(f.read(), Name=attachment_filename)
            attach['Content-Disposition'] = f'attachment; filename="{attachment_filename}"'
            msg.attach(attach)

    # Send the email
    with smtplib.SMTP_SSL(smtp_address, smtp_port) as server:
        server.login(send_email_address, send_email_password)
        server.sendmail(send_email_address, receive_email_address, msg.as_string())

def local_to_df(path, partial_file_name, skip_rows=0, keep_file_name=False, sheet_num=1, encoding='auto', sep=','):
    all_data = pd.DataFrame()

    # 获取匹配的文件列表，并为其添加进度条
    file_list = list(glob.glob(f"{path}/**/*{partial_file_name}*.*", recursive=True))
    
    for file_name in tqdm(file_list, desc="Processing Files"):
        _, file_extension = os.path.splitext(file_name)
        file_extension = file_extension[1:]  # remove the leading dot
        
        # 识别文件编码类型
        if encoding=='auto':
            with open(file_name, 'rb') as file:
                result = chardet.detect(file.read())
                encoding = result['encoding']
        
        if file_extension == "csv":
            file_data = pd.read_csv(file_name, encoding=encoding, skiprows=skip_rows, sep=sep)
        elif file_extension == "xlsx":
            file_data = pd.read_excel(file_name, sheet_name=sheet_num - 1, skiprows=skip_rows)
        else:
            continue  # Skip unsupported file types

        if keep_file_name:
            file_data["file_name"] = str(file_name)

        all_data = pd.concat([all_data, file_data])

    if all_data.empty:
        print("No matching file in the path or empty data found")
        return None
    return all_data

def df_to_xlsx(df, directory_path, file_name):
    os.makedirs(directory_path, exist_ok=True)
    file_name = f"{file_name}.xlsx" if not file_name.endswith('.xlsx') else file_name
    df.to_excel(os.path.join(directory_path, file_name), index=False)

def df_to_csv(df, directory_path, file_name):
    os.makedirs(directory_path, exist_ok=True)
    file_name = f"{file_name}.csv" if not file_name.endswith('.csv') else file_name
    df.to_csv(os.path.join(directory_path, file_name), index=False)


import numpy as np

def check_empty(df, columns=None, output_cols=None):
    if columns is not None and not isinstance(columns, list):
        columns = [columns]
    # 创建临时DataFrame
    tmp_df = df[columns].copy() if columns else df.copy()

    # 替换空白字符串和None为 np.nan
    tmp_df.replace(r'^\s*$', np.nan, regex=True, inplace=True)
    tmp_df.fillna(value=np.nan, inplace=True)

    # 计算每列的缺失值数量
    missing_counts = tmp_df.isnull().sum().sort_values(ascending=False)

    # 总行数
    total_rows = len(df)

    # 可视化缺失数据占比
    if missing_counts.any():
        print("缺失数据：")
        for col, count in missing_counts.items():
            num_chars = int((count / total_rows) * 10)  # 根据比例计算字符数量
            print(f"{col}: {'█' * num_chars}{'.' * (10 - num_chars)} ({count} rows)")
        null_data = df[tmp_df.isnull().any(axis=1)]
        if output_cols: 
            null_data = null_data[output_cols].drop_duplicates()
        return null_data
    else:
        print("DataFrame 数据完整。")


import setuptools

with open("README.md", "r") as fh:
  long_description = fh.read()

setuptools.setup(
  name="datarecipe",
  version="2.0.3",
  author="Hanfan Chen",
  author_email="HanfanC@outlook.com",
  description="A recipe for every data baker",
  long_description=long_description,
  long_description_content_type='text/markdown',
  url="https://github.com/Adward-Chen/databaker/tree/main",
  packages=setuptools.find_packages(),
  classifiers=[
  "Programming Language :: Python :: 3.10",
  "License :: OSI Approved :: MIT License",
  "Operating System :: OS Independent",
  ],
)